

#include "galil1800.h"


struct pci_device_id galil1800_idtable[] = {
	{  .vendor = PLX_PCI_VENDOR_ID,
	   .device = PLX_PCI_DEVICE_ID,
	   .subvendor = GALIL_SUB_VENDOR,
	   .subdevice = GALIL_SUB_DEVICE_1800,
	   .class = 0,
	   .class_mask = 0,
	   .driver_data = 0 },
	{ 0, }
};

int galil1800_probe(struct pci_dev *pcidev,
			const struct pci_device_id *devid) __init;

void galil1800_remove(struct pci_dev *pcidev) __exit;

struct pci_driver galil1800_driver  = {
	.name = GALIL1800_MODULE_NAME,
	.id_table = galil1800_idtable,
	.probe = galil1800_probe,
	.remove = __exit_p(galil1800_remove),
	.suspend = NULL,
	.resume = NULL,
	.enable_wake = NULL,
	.shutdown = NULL
};


MODULE_AUTHOR("Universit di Roma Tor Vergata (E. Betti)");
MODULE_DESCRIPTION("Galil DMC 1800 Example Driver for"
		" Free Course of Linux Kernel Hacking 2006");
MODULE_VERSION("0.1");
MODULE_ALIAS("galil1800");
MODULE_LICENSE("GPL");


/* init module function */
int __init galil1800_init(void)
{
	int rc;

	GALIL_PRINT_DEBUG("additional debug print enabled\n");

	rc = pci_register_driver(&galil1800_driver);
	if (rc) {
		printk(KERN_ERR PFX "cannot register the"
			" PCI device driver\n");
		return rc;
	}

	GALIL_PRINT_DEBUG("module loaded\n");
	return 0;

}

/* cleanup module function */
void __exit galil1800_cleanup(void)
{
	pci_unregister_driver(&galil1800_driver);

	GALIL_PRINT_DEBUG("module unloaded\n");
}

module_init(galil1800_init);
module_exit(galil1800_cleanup);


/* "Probe" function: it runs one time for each device detected */
int __init galil1800_probe(struct pci_dev *pcidev,
			const struct pci_device_id *devid)
{
	int ret;
	struct galil1800_state *s;

	s = kzalloc(sizeof(struct galil1800_state), GFP_KERNEL);
	if (!s){
		printk(KERN_ERR PFX "alloc of galil1800_state"
					" struct failed\n");
		return -ENOMEM;
	}
	pci_set_drvdata(pcidev, s);

	if (pcidev->irq == 0) {
		printk(KERN_ERR PFX "IRQ in pci_dev is 0\n");
		ret = -1;
		goto out1;
	}

	if(pci_resource_len(pcidev, GALIL1800_REG_BAR_INDEX) <
				GALIL1800_REG_PCI_RESOURCE_LEN) {
		printk(KERN_ERR PFX "I/O ports size is not %d\n",
				GALIL1800_REG_PCI_RESOURCE_LEN);
		ret = -1;
		goto out1;
	}

	s->mem_size = pci_resource_len(pcidev,
					GALIL1800_DPRAM_BAR_INDEX);
	if(!s->mem_size) {
		printk(KERN_ERR PFX "I/O memory size is 0\n");
		ret = -1;
		goto out1;
	}


	s->io = pci_resource_start(pcidev, GALIL1800_REG_BAR_INDEX);
	if (!request_region(s->io, GALIL1800_REG_PCI_RESOURCE_LEN,
				GALIL1800_MODULE_NAME) ) {
		printk(KERN_ERR PFX "I/O ports %#lx->%#lx in use\n",
		  s->io, s->io + GALIL1800_REG_PCI_RESOURCE_LEN - 1);
		ret = -1;
		goto out1;
	}

	s->mem_paddr = pci_resource_start(pcidev,
					GALIL1800_DPRAM_BAR_INDEX);
	if (!request_mem_region(s->mem_paddr, s->mem_size,
				GALIL1800_MODULE_NAME)) {
		printk(KERN_ERR PFX "I/O memory %#lx->%#lx in use\n",
			s->mem_paddr, s->mem_paddr + s->mem_size-1);
		ret = -1;
		goto out2;
	}

	if (!(s->mem_vaddr = ioremap(s->mem_paddr, s->mem_size))) {
		printk(KERN_ERR PFX "ioremap failed");
		ret = -1;
		goto out3;
	}


	s->ctrl_reg  = s->io + 0x4;
	s->irq_reg   = s->io + 0x8;
	s->chan2_reg = s->io + 0xC;
	s->irq = pcidev->irq;
	spin_lock_init(&(s->hwlock));

	ret = pci_enable_device(pcidev);
	if (ret) {
		printk(KERN_ERR PFX "cannot enable PCI device\n");
		goto out4;
	}

	ret = galil1800_init_controller_probe_step(s);
	if(ret) {
		printk(KERN_ERR PFX "failed to initialize the device\n");
		goto out5;
	}

	printk(KERN_INFO PFX "base I/O port 0x%lx, "
		"I/O memory p:0x%lx v:0x%lx"
		" (size %lu), IRQ %d\n", s->io, s->mem_paddr,
		(unsigned long)s->mem_vaddr, s->mem_size,
		pcidev->irq);

	ret = alloc_chrdev_region(&s->dev, 0, 1,
				GALIL1800_MODULE_NAME);
	if (ret < 0) {
		printk(KERN_ERR PFX
			"cannot register the device files\n");
		goto out5;
	}
	GALIL_PRINT_DEBUG("major is %d and minor is %d\n",
			MAJOR(s->dev), MINOR(s->dev));

	cdev_init(&s->cdev, &galil1800_fops);
	s->cdev.owner = THIS_MODULE;

	ret=cdev_add(&s->cdev, s->dev, 1);
	if (ret) {
		printk(KERN_INFO PFX
			"cannot add cdev\n");
		goto out6;
	}

	GALIL_PRINT_DEBUG("galil1800_probe: ended\n");
	return 0;
out6:
	unregister_chrdev_region(s->dev, 1);
out5:
	pci_disable_device(pcidev);
out4:
	iounmap(s->mem_vaddr);
out3:
	release_mem_region(s->mem_paddr, s->mem_size);
out2:
	release_region(s->io, GALIL1800_REG_PCI_RESOURCE_LEN);
out1:
	pci_set_drvdata(pcidev, NULL);
	kfree(s);
	return ret;
}

void __exit galil1800_remove(struct pci_dev *pcidev)
{
	struct galil1800_state *s = pci_get_drvdata(pcidev);
	cdev_del(&s->cdev);
	unregister_chrdev_region(s->dev, 1);
	pci_disable_device(pcidev);
	iounmap(s->mem_vaddr);
	release_region(s->io, GALIL1800_REG_PCI_RESOURCE_LEN);
	release_mem_region(s->mem_paddr, s->mem_size);
	pci_set_drvdata(pcidev, NULL);
	kfree(s);
	GALIL_PRINT_DEBUG("galil1800_remove: ended\n");
}
