
#ifndef _GALIL1800_IO_H_
#define _GALIL1800_IO_H_

#include "galil1800.h"

#include <asm/io.h>
#include <linux/sched.h>
#include <linux/delay.h>

/* jiffies for [read,write] timeout. (50ms) */
#define GALIL1800_TIMEOUT 50*HZ/1000


#define GALIL1800_READ_FIFO_SIZE    512
#define GALIL1800_WRITE_FIFO_SIZE   512

/* max byte that you can write,
   if GALIL1800_FIFO_HALF_FLAG is not set */
#define GALIL1800_HALF_FULL_WRITE_BURST 255


#define GALIL1800_READ_REG(_S)         ((_S)->io)
#define GALIL1800_WRITE_REG(_S)        GALIL1800_READ_REG(_S)
#define GALIL1800_CONTROL_REG(_S)      ((_S)->ctrl_reg)
#define GALIL1800_IRQ_REG(_S)          ((_S)->irq_reg)
#define GALIL1800_RESET_REG(_S)        GALIL1800_IRQ_REG(_S)
#define GALIL1800_2CHAN_REG(_S)        ((_S)->chan2_reg)


/* FIFO Control register values (size 8 bit) */
#define GALIL1800_FIFO_FULL_FLAG      (0x01)
#define GALIL1800_FIFO_HALF_FLAG      (0x02)
#define GALIL1800_FIFO_EMPTY_FLAG     (0x04)
#define GALIL1800_FIFO2_BUSY_FLAG     (0x08)
#define GALIL1800_FIFO2_FREEZE_FLAG   (0x10)
#define GALIL1800_IRQ_STATUS_FLAG     (0x20)
#define GALIL1800_IRQ_ENABLE_FLAG     (0x40)
#define GALIL1800_FIFO2_EMPTY_FLAG    (0x80)


/* Reset register values (size 8 bit)
   !! Don't use "OR" whit this flag, otherwise the reset don't work !! */
#define __GALIL1800_RESET_FIFO_R    0x02
#define __GALIL1800_RESET_FIFO_W    0x04
#define __GALIL1800_RESET_HARDWARE  0x80

/* Reset timeout in msec */
#define GALIL1800_RESET_TIMEOUT 100

/* Reset the controller: use macro to select what you reset */
static inline void __galil1800_reset(struct galil1800_state *s, char reset)
{
	unsigned long flags;
	spin_lock_irqsave(&s->hwlock, flags);
	outb(reset, GALIL1800_RESET_REG(s));
	spin_unlock_irqrestore(&s->hwlock, flags);
	schedule_timeout_uninterruptible(GALIL1800_RESET_TIMEOUT);
}

#define galil1800_reset_fifo_r(_S)     \
		__galil1800_reset(_S, __GALIL1800_RESET_FIFO_R)

#define galil1800_reset_fifo_w(_S)     \
		__galil1800_reset(_S, __GALIL1800_RESET_FIFO_W)

#define galil1800_reset_fifo_rw(_S)          \
	do {                                 \
		galil1800_reset_fifo_r(_S);  \
		galil1800_reset_fifo_w(_S);  \
	}while(0)

#define galil1800_reset_hardware(_S)      \
	__galil1800_reset(_S, __GALIL1800_RESET_HARDWARE)


static inline void galil1800_enable_IRQ(struct galil1800_state *s)
{
	unsigned long flags;
	spin_lock_irqsave(&s->hwlock, flags);
	inb(GALIL1800_IRQ_REG(s)); /* useless ?! */
	outb(inb(GALIL1800_CONTROL_REG(s)) | GALIL1800_IRQ_ENABLE_FLAG,
						GALIL1800_CONTROL_REG(s));
	inb(GALIL1800_IRQ_REG(s)); /* useless ?! */
	spin_unlock_irqrestore(&s->hwlock, flags);
}

static inline void galil1800_disable_IRQ(struct galil1800_state *s)
{
	unsigned long flags;
	spin_lock_irqsave(&s->hwlock, flags);
	outb(inb(GALIL1800_CONTROL_REG(s)) & ~GALIL1800_IRQ_ENABLE_FLAG,
						GALIL1800_CONTROL_REG(s));
	spin_unlock_irqrestore(&s->hwlock, flags);
}

/* Check for a byte in the Read FIFO
   Return 1 if a byte is available in the FIFO, 0 otherwise. */
static inline int __galil1800_may_read(struct galil1800_state *s)
{
	return !(inb(GALIL1800_CONTROL_REG(s)) & GALIL1800_FIFO_EMPTY_FLAG);
}

/* Wait for a byte in the Read FIFO, or until a timeout occurs.
   Return 1 if a byte is available in the FIFO, 0 otherwise.
   Interrupts MUST be enabled, or the kernel will freeze */
static inline int galil1800_may_read(struct galil1800_state *s)
{
	unsigned long timeout = jiffies + GALIL1800_TIMEOUT;
	do {
		if (__galil1800_may_read(s))
			return 1;
	} while (time_before(jiffies,timeout));
	return 0;
}

/* Read from the read fifo */
extern ssize_t galil1800_recv(struct galil1800_state *, char *, ssize_t);

/* Check for a free slot in the Write FIFO, without delay.
   Return:
            2 if FIFO is less than half full (at least 255 free slots),
            1 if a slot is available in the FIFO,
            0 otherwise. */
static inline int __galil1800_may_write(struct galil1800_state *s)
{
	char t = inb(GALIL1800_CONTROL_REG(s));
	if(!(t & GALIL1800_FIFO_HALF_FLAG))
		return 2;
	if(!(t & GALIL1800_FIFO_FULL_FLAG))
		return 1;
	return 0;
}


/* Wait for a free slot in the Write FIFO, or until a timeout occurs.
   Return:
            2 if FIFO is less than half full (at least 255 free slots),
            1 if a slot is available in the FIFO,
            0 otherwise.
   Local interrupts MUST be enabled */
static inline int galil1800_may_write(struct galil1800_state *s)
{
	unsigned long timeout = jiffies + GALIL1800_TIMEOUT;
	char t;
	do {
		t = inb(GALIL1800_CONTROL_REG(s));
		if (!(t & GALIL1800_FIFO_HALF_FLAG))
			return 2;
		if (!(t & GALIL1800_FIFO_FULL_FLAG))
			return 1;
	} while (time_before(jiffies,timeout));
	return 0;
}

/* Write data in the write fifo */
extern ssize_t galil1800_send(struct galil1800_state *, const char *, size_t);


#endif
