
#include <asm/uaccess.h>
#include <linux/fs.h>

#include "galil1800_io.h"


static
ssize_t galil1800_write(struct file *filp,
	const char __user *buf, size_t size, loff_t *poff)
{
	struct galil1800_state *s =
		(struct galil1800_state *)filp->private_data;
	char *_buf;
	ssize_t request;
	int rc=0;

	GALIL_PRINT_DEBUG("write called\n");

	if (!size)
		return 0;
	if (!buf || size < 0)
		return -EINVAL;

	request = (size < GALIL1800_WRITE_FIFO_SIZE ?
			size : GALIL1800_WRITE_FIFO_SIZE);

	_buf = kmalloc(request, GFP_KERNEL);
	if (!_buf) {
		printk(KERN_ERR PFX "write: alloc _buf failed\n");
		return -ENOMEM;
	}

	if (copy_from_user(_buf, buf, request)) {
		rc = -EFAULT;
		goto out1;
	}

	rc = galil1800_send(s, _buf, request);

out1:
	kfree(_buf);
	return rc;
}

static
ssize_t galil1800_read(struct file *filp, char __user *buf,
				size_t size, loff_t *poff)
{
	struct galil1800_state *s =
		(struct galil1800_state *) filp->private_data;
	char *_buf;
	ssize_t count;
	ssize_t request = (size < GALIL1800_READ_FIFO_SIZE ?
					size : GALIL1800_READ_FIFO_SIZE);
	int rc = 0;
	unsigned long timeout = 50*HZ/1000;

	GALIL_PRINT_DEBUG("read called\n");

	/* Wait to data arrive */
	while(__galil1800_may_read(s)==0) {
		if(filp->f_flags & O_NONBLOCK)
			return -EAGAIN;

		GALIL_ONE_PRINT_DEBUG("read going to sleep"
					" (read fifo empty)\n");

		rc = wait_event_interruptible_timeout(s->read_wait,
					__galil1800_may_read(s)>0, timeout);
		if(rc == -ERESTARTSYS) {
			GALIL_PRINT_DEBUG("read interrupted\n");
			return rc;
		}
	}

	_buf = kmalloc(request, GFP_KERNEL);
	if (!_buf) {
		printk(KERN_ERR PFX "read: alloc _buf failed\n");
		return -ENOMEM;
	}

	count = galil1800_recv(s, _buf, request);

	if (count > 0) {
		rc = copy_to_user(buf, _buf, count);
		rc = rc > 0 ? -EFAULT : count;
	}

	kfree(_buf);
	return rc;
}

static
int galil1800_open(struct inode *inode, struct file *filp)
{
	struct galil1800_state *s;

	GALIL_PRINT_DEBUG("open called\n");

	s = container_of(inode->i_cdev,
		struct galil1800_state, cdev);
	filp->private_data = s;

	if(down_interruptible(&s->open_sem)) {
		GALIL_PRINT_DEBUG("\"open\": \"down\" interrupted\n");
		return -ERESTARTSYS;
	}

	/* Whit this type of check,
	   only one process can open for read and another for write */
	while (s->open_mode & filp->f_mode) {
		if (filp->f_flags & O_NONBLOCK) { /* O_NONBLOCK == O_NDELAY */
			up(&s->open_sem);
			GALIL_PRINT_DEBUG("\"open\":"
				" O_NONBLOCK is set and device is busy\n");
			return -EAGAIN;
		}
		up(&s->open_sem);
		if(wait_event_interruptible(s->open_wait,
			!(s->open_mode & filp->f_mode) ) ) {
			GALIL_PRINT_DEBUG("\"open\":"
					" \"wait_event\" interrupted\n");
			return -ERESTARTSYS;
		}
		if(down_interruptible(&s->open_sem)) {
			GALIL_PRINT_DEBUG("\"open\":"
					" \"down\" interrupted\n");
			return -ERESTARTSYS;
		}
	}

	if(!s->open_mode) {
		int ret;
		ret = galil1800_init_controller_open_step(s);
		if(ret) {
			up(&s->open_sem);
			return ret;
		}
	}

	s->open_mode |= filp->f_mode & (FMODE_READ | FMODE_WRITE);

	up(&s->open_sem);

	nonseekable_open(inode,filp);

	return 0;
}

static
int galil1800_release(struct inode *inode, struct file *filp)
{
	struct galil1800_state *s =
		(struct galil1800_state *)filp->private_data;

	GALIL_PRINT_DEBUG("release called\n");

	if(down_interruptible(&s->open_sem)) {
		GALIL_PRINT_DEBUG("\"release\":"
				" \"down\" interrupted\n");
		return -ERESTARTSYS;
	}

	s->open_mode &= (~filp->f_mode) & (FMODE_READ | FMODE_WRITE);

	if(!s->open_mode)
		galil1800_clean_controller_open_step(s);

	up(&s->open_sem);

	wake_up_interruptible_sync(&s->open_wait);

	return 0;
}


struct file_operations galil1800_fops = {
	.owner = THIS_MODULE,
	.read = galil1800_read,
	.write = galil1800_write,
	.open = galil1800_open,
	.release = galil1800_release,
	.llseek = no_llseek
};
